/*-
 * Copyright (c) 2012 Michael Tuexen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#define DATA_BUFFER_SIZE (1<<10)
#define CONTROL_BUFFER_SIZE (1<<10)

int
main(void)
{
	int fd;
	ssize_t n;
	struct sockaddr_in addr;
	struct cmsghdr *cmsg;
	struct msghdr msg;
	struct iovec iov;
	const int on = 1;
	char dbuf[DATA_BUFFER_SIZE];
	char cbuf[CONTROL_BUFFER_SIZE];
	char nbuf[INET_ADDRSTRLEN];
	const char *name;

	if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
	}

	if (setsockopt(fd, IPPROTO_IP, IP_RECVTOS, &on, sizeof(on)) < 0) {
		perror("IP_RECVTOS");
	}
	if (setsockopt(fd, IPPROTO_IP, IP_RECVTTL, &on, sizeof(on)) < 0) {
		perror("IP_RECVTTL");
	}
	if (setsockopt(fd, IPPROTO_IP, IP_RECVDSTADDR, &on, sizeof(on)) < 0) {
		perror("IP_RECVDSTADDR");
	}

	memset(&addr, 0, sizeof(struct sockaddr_in));
	addr.sin_family = AF_INET;
#ifdef HAVE_SIN_LEN
	addr.sin_len = sizeof(struct sockaddr_in);
#endif
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(9);
	if (bind(fd, (struct sockaddr*) &addr, sizeof(struct sockaddr_in)) < 0) {
		perror("bind");
	}

	for (;;) {
		iov.iov_base = dbuf;
		iov.iov_len = DATA_BUFFER_SIZE;
		memset(&msg, 0, sizeof(struct msghdr));
		msg.msg_name = &addr;
		msg.msg_namelen = sizeof(struct sockaddr_in6);
		msg.msg_iov = &iov;
		msg.msg_iovlen  = 1;
		msg.msg_control = cbuf;
		msg.msg_controllen = CONTROL_BUFFER_SIZE;
		n = recvmsg(fd, &msg, 0);
		if (n < 0) {
			perror("recvmsg");
			break;
		}
		name = inet_ntop(AF_INET, &addr.sin_addr, nbuf, INET_ADDRSTRLEN);
		printf("Received message of length %ld from %s.\n", n, name);
		for (cmsg = CMSG_FIRSTHDR(&msg); cmsg != NULL; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
			printf("level = %d, type = %d, len = %d.\n",
			       cmsg->cmsg_level, cmsg->cmsg_type, cmsg->cmsg_len);
			if ((cmsg->cmsg_level == IPPROTO_IP) &&
			    (cmsg->cmsg_type == IP_RECVTOS) &&
			    (cmsg->cmsg_len == CMSG_LEN(sizeof(uint8_t)))) {
				uint8_t *tos;

				tos = (uint8_t *)CMSG_DATA(cmsg);
				printf("Received tos = %d\n", *tos);
			}
			if ((cmsg->cmsg_level == IPPROTO_IP) &&
			    (cmsg->cmsg_type == IP_RECVTTL) &&
			    (cmsg->cmsg_len == CMSG_LEN(sizeof(uint8_t)))) {
				uint8_t *ttl;

				ttl = (uint8_t *)CMSG_DATA(cmsg);
				printf("Received TTL = %d\n", *ttl);
			}
			if ((cmsg->cmsg_level == IPPROTO_IP) &&
			    (cmsg->cmsg_type == IP_RECVDSTADDR) &&
			    (cmsg->cmsg_len == CMSG_LEN(sizeof(struct in_addr)))) {
				struct in_addr *dst;

				dst = (struct in_addr *)CMSG_DATA(cmsg);
				name = inet_ntop(AF_INET, dst, nbuf, INET_ADDRSTRLEN);
				printf("Destination address %s.\n", name);
			}
		}
	}
	if (close(fd) < 0) {
		perror("close");
	}
	return (0);
}
